//
// Created by Kai Zhao on 6/12/20.
//

#ifndef SZ3_PREPROCESSOR_HPP
#define SZ3_PREPROCESSOR_HPP

#include "SZ3/def.hpp"

namespace SZ3 {
    namespace concepts {

        template<class T, uint N>
        class PreprocessorInterface {
        public:
//            virtual void preprocess(T *data, std::array <size_t, N> dims) = 0;
        };
    }
}

#endif //SZ_PREPROCESSOR_HPP
